/**
 * JIST Extensions for Computer-Integrated Surgery
 *
 * Center for Computer-Integrated Surgical Systems and Technology &
 * Johns Hopkins Applied Physics Laboratory &
 * The Johns Hopkins University
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.  The license is available for reading at:
 * http://www.gnu.org/copyleft/lgpl.html
 *
 * @author Blake Lucas
 */
package edu.jhu.cs.cisst.plugins.visualization;

import java.awt.Dimension;
import java.io.File;

import processing.core.PImage;
import edu.jhu.cs.cisst.jist.pipeline.io.PImageReaderWriter;
import edu.jhu.cs.cisst.vent.Visualization;
import edu.jhu.cs.cisst.vent.VisualizationApplication;
import edu.jhu.cs.cisst.vent.VisualizationPlugIn;
import edu.jhu.cs.cisst.vent.VisualizationProcessing;
import edu.jhu.cs.cisst.vent.widgets.VisualizationTriangleMesh;
import edu.jhu.ece.iacl.algorithms.CommonAuthors;
import edu.jhu.ece.iacl.jist.io.FileExtensionFilter;
import edu.jhu.ece.iacl.jist.io.FileReaderWriter;
import edu.jhu.ece.iacl.jist.pipeline.AlgorithmInformation;
import edu.jhu.ece.iacl.jist.pipeline.AlgorithmRuntimeException;
import edu.jhu.ece.iacl.jist.pipeline.CalculationMonitor;
import edu.jhu.ece.iacl.jist.pipeline.DevelopmentStatus;
import edu.jhu.ece.iacl.jist.pipeline.ProcessingAlgorithm;
import edu.jhu.ece.iacl.jist.pipeline.parameter.ParamBoolean;
import edu.jhu.ece.iacl.jist.pipeline.parameter.ParamCollection;
import edu.jhu.ece.iacl.jist.pipeline.parameter.ParamFile;
import edu.jhu.ece.iacl.jist.pipeline.parameter.ParamInteger;
import edu.jhu.ece.iacl.jist.pipeline.parameter.ParamObject;
import edu.jhu.ece.iacl.jist.pipeline.parameter.ParamSurfaceCollection;
import edu.jhu.ece.iacl.jist.structures.geom.EmbeddedSurface;

// TODO: Auto-generated Javadoc
/**
 * The Class PlugInVisualizeSurface.
 */
public class PlugInVisualizeSurface extends ProcessingAlgorithm implements
		VisualizationPlugIn {

	/** The surf param. */
	protected ParamSurfaceCollection surfParam;
	/** The image height param. */
	ParamInteger imageWidthParam, imageHeightParam;
	protected ParamFile cacheParam;
	protected ParamBoolean captureParam;
	protected ParamObject<PImage> screenshotParam;

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.jhu.ece.iacl.jist.pipeline.ProcessingAlgorithm#createInputParameters
	 * (edu.jhu.ece.iacl.jist.pipeline.parameter.ParamCollection)
	 */
	protected void createInputParameters(ParamCollection inputParams) {
		inputParams.add(surfParam = new ParamSurfaceCollection("Surfaces"));
		inputParams
				.add(imageWidthParam = new ParamInteger("Image Width", 1024));
		inputParams
				.add(imageHeightParam = new ParamInteger("Image Height", 768));
		inputParams.add(captureParam = new ParamBoolean(
				"Only Capture Screenshot", false));
		inputParams.add(cacheParam = new ParamFile("Visualization Parameters",
				new FileExtensionFilter(new String[] { "cache" })));
		cacheParam.setMandatory(false);

		inputParams.setName("vis_surfaces");
		inputParams.setLabel("Visualize Surfaces");
		inputParams.setPackage("CISST");
		inputParams.setCategory("Visualization");

		AlgorithmInformation info = getAlgorithmInformation();
		info.add(CommonAuthors.blakeLucas);
		info.setAffiliation("Johns Hopkins University, Department of Computer Science");
		info.setDescription("Visualizes surfaces.");
		info.setVersion(VisualizationTriangleMesh.getVersion());
		info.setStatus(DevelopmentStatus.RC);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.jhu.ece.iacl.jist.pipeline.ProcessingAlgorithm#createOutputParameters
	 * (edu.jhu.ece.iacl.jist.pipeline.parameter.ParamCollection)
	 */
	@Override
	protected void createOutputParameters(ParamCollection outputParams) {
		outputParams.add(screenshotParam = new ParamObject<PImage>(
				"Screenshot", new PImageReaderWriter()));
		screenshotParam.setMandatory(false);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * edu.jhu.ece.iacl.jist.pipeline.ProcessingAlgorithm#execute(edu.jhu.ece
	 * .iacl.jist.pipeline.CalculationMonitor)
	 */
	@Override
	protected void execute(CalculationMonitor monitor)
			throws AlgorithmRuntimeException {

		Visualization visual = createVisualization();
		VisualizationApplication app = new VisualizationApplication(this,
				visual);
		app.setMinimumSize(new Dimension(imageWidthParam.getInt() + 400,
				imageHeightParam.getInt() + 50));
		screenshotParam.setFileName(FileReaderWriter.getFileName(surfParam
				.getValue().get(0)));
		if (captureParam.getValue()) {
			File cacheFile = cacheParam.getValue();

			screenshotParam.setObject(app.capture(null, cacheFile));
			//visual.dispose();
		} else {
			app.runAndWait();
			PImage img = new PImage(visual.getScreenshot());
			screenshotParam.setObject(img);
			//visual.dispose();
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.jhu.cs.cisst.vent.VisualizationPlugIn#createVisualization()
	 */
	@Override
	public Visualization createVisualization() {
		VisualizationTriangleMesh visual = new VisualizationTriangleMesh(1024,
				768);
		for (EmbeddedSurface surf : surfParam.getSurfaceList()) {
			visual.add(surf);
		}
		return visual;
	}

}
